#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>

#include "xhkeys_conf.h"

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif


#include "xhkeys.h"

extern Display *dpy;
extern Window wnd;
extern Atom WMStateAtom;


const char *SpecCodeNames[] = {
#ifdef CDROM_SUPPORT
        "CDPlayPause", "CDPlay", 
		"CDPause",     "CDStop",
	    "CDLoadEject", "CDLoad",
		"CDEject",     "CDSlot", 
#endif
#ifdef MIXER_SUPPORT
	    "SndVolume", "SndBalance", 
#endif
	    "WndRotate", "Quit" };

const int SpecCodeCount = sizeof(SpecCodeNames) / sizeof(char *);

// Dir:  >=0 bottom to top,  < 0 top to bottom

Window lookupWindow(Window cur, int dir, LookupWindowHandler handler, const void *par)
{ 
  Window tmpwnd, *children;
  unsigned int nchildren;
  int i, j;

  if ((*handler)(cur, par)) return cur;
  

  if (XQueryTree(dpy, cur, &tmpwnd, &tmpwnd, &children, &nchildren) == False)
  {	 children = NULL; nchildren = 0; }

  cur = None;
  for (i=0; i<nchildren; i++) {
    j = (dir >= 0) ? i : nchildren-i-1;
    if ((cur = lookupWindow(children[j], dir, handler, par)) != None)
		return cur;
  }  	

  if (children)	XFree(children);

  return cur;
}


//====================================================================

static Bool WndByWMClass(Window cur, const XClassHint *xch_needed)
{
  XClassHint   xch;
  
  return XGetClassHint(dpy, cur, &xch) &&
		 (strcmp(xch.res_name, xch_needed->res_name) == 0) &&	  
		 (strcmp(xch.res_class, xch_needed->res_class) == 0);
}


Window findWindowByWMClass(Window main_wnd,
                           const char *app_name, const char *class_name)
{
  XClassHint xch;
  
  xch.res_name = (char *) app_name;
  xch.res_class = (char *) class_name;
  
  return lookupWindow(main_wnd, -1, (LookupWindowHandler) WndByWMClass, &xch);

}

static int wmClassCount;

static Bool WndCountWMClass(Window cur, const XClassHint *xch_needed)
{
  XClassHint   xch;
  
  if ( XGetClassHint(dpy, cur, &xch) &&
	   (strcmp(xch.res_name, xch_needed->res_name) == 0) &&	  
	   (strcmp(xch.res_class, xch_needed->res_class) == 0)
	 ) wmClassCount++;
		 
  return False;
}


int countWindowsByWMClass(Window main_wnd,
                           const char *app_name, const char *class_name)
{
  XClassHint xch;
  
  xch.res_name = (char *) app_name;
  xch.res_class = (char *) class_name;
  wmClassCount = 0;  

  lookupWindow(main_wnd, -1, (LookupWindowHandler) WndCountWMClass, &xch);
  return wmClassCount;
}

//====================================================================
static Bool WndById(Window cur, Window id_needed)
{
  return (cur == id_needed);
}


Window findWindowById(Window main_wnd, Window id_needed)
{
  return lookupWindow(main_wnd, -1, (LookupWindowHandler) WndById, (const void *)id_needed);

}

//====================================================================
static Bool WndByTitle(Window cur, const char *title_needed)
{
  XTextProperty xtp;
  char **list;
  int list_count;
  Bool result = False;
  
  		
  if (XGetWMName(dpy, cur, &xtp) &&
	  XTextPropertyToStringList(&xtp, &list, &list_count)) {
	  result = (list_count > 0 && strcmp(list[0], title_needed) == 0); 
	  XFreeStringList(list);		
  }  	 

  return result;
}


Window findWindowByTitle(Window main_wnd,  const char *title)
{
  return lookupWindow(main_wnd, -1, (LookupWindowHandler) WndByTitle, title);

}
//====================================================================

Bool  GetWindowState(Window window, long *state)
{
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *data;
	Bool success;

    XGetWindowProperty(dpy, window, WMStateAtom, 0,
                             2, False, WMStateAtom,
                             &actual_type,  &actual_format,
                             &nitems, &bytes_after, &data);

	success = (actual_type == WMStateAtom && nitems > 0);
//	success = (actual_type != None && nitems > 0);
    if (success) *state = *(long *)data; 
    XFree(data);
    return success;
}

static Bool IsVisibleWindow(Window cur, const void *dummy )
{
  long state;
  
  return GetWindowState(cur, &state) && (state >= 0);
  
}

// Find top window having WM_STATE = state
Window findVisibleWindow(Window main_wnd, int dir)
{
  long state;
  return lookupWindow(main_wnd, dir, IsVisibleWindow, &state);

}


static Bool IsClientWindow(Window cur, const void *dummy )
{
  long state;
  return GetWindowState(cur, &state);
}


// Find top window that has WM_STATE property
Window findClientWindow(Window main_wnd, int dir)
{
  return lookupWindow(main_wnd, dir, IsClientWindow, NULL);

}


#ifndef HAVE_STRSEP
char *strsep(char **stringp, const char *delim)
{
  char *start, *sep;;

  start = *stringp;
  sep = strpbrk(start, delim);
  if (sep != NULL) *sep++ = '\0';
  *stringp = sep;

  return start;
}
#endif
